import shutil
import subprocess  # nosec

import typer


def run(cmd: list[str]):
    orig = cmd[0]
    cmd[0] = shutil.which(orig)
    if cmd[0] is None:
        raise subprocess.SubprocessError(f"Invalid executable {orig}")

    result = subprocess.run(cmd, capture_output=True)  # nosec
    if result.returncode:
        typer.echo(typer.style(result.stdout.decode("utf-8"), fg=typer.colors.GREEN))
        typer.echo(typer.style(result.stderr.decode("utf-8"), fg=typer.colors.RED))
        raise subprocess.CalledProcessError(result.returncode, cmd)

    return result



import re
from datetime import datetime, timedelta, timezone

import typer

from gcrc import gcloud
from gcrc.gcloud import ImageTag
from gcrc.settings import conf

DT_FORMAT = "%Y-%m-%d"
KEEP_EXTRAS: dict[str, re.Pattern] = {val: re.compile(val) for val in conf.KEEP_EXTRA}
_BEFORE = None
app = typer.Typer()


def _get_before():
    global _BEFORE
    if not _BEFORE:
        _BEFORE = datetime.now(timezone.utc) - timedelta(days=-conf.KEEP_TAGS_DAYS)

    return _BEFORE


def _pick_cleanup_tags(tags: list[ImageTag]) -> (list[ImageTag], list[ImageTag]):
    """
    Figure out which tags to keep and which to clean up, based on

    1) age - "KEEP_TAGS_DAYS"
    2) important tags - "KEEP_EXTRA"
    3) minimum limits - "KEEP_TAGS_MIN"

    Will not always clean up everything that could be cleaned up, but
    guarantees the desired things are always kept.
    """
    keep = []
    clean = []
    extras = {}
    before = _get_before()

    # Process latest first
    for tag in sorted(tags, key=lambda t: t.timestamp, reverse=True):
        keep_this = False

        # Keep recent items
        if tag.timestamp >= before:
            keep_this = True

        # Keep extras
        for key, regex in KEEP_EXTRAS.items():
            for t in tag.tags:
                if regex.match(t):
                    extras[key] = extras.get(key, []) + [tag]
                    if len(extras[key]) <= conf.KEEP_TAGS_MIN:
                        keep_this = True

        # Keep minimum X items
        if len(keep) < conf.KEEP_TAGS_MIN:
            keep_this = True

        if keep_this:
            keep.append(tag)
        else:
            clean.append(tag)

    return keep, clean


@app.command()
def list_images(gcr_repo: str):
    """
    List information about all the images on the repository
    """
    typer.echo(
        typer.style(
            f" ----- {gcr_repo} images ----- ", fg=typer.colors.GREEN, bold=True
        )
    )
    for image in gcloud.list_images(gcr_repo):
        typer.echo(image)


@app.command()
def image_info(gcr_repo: str):
    """
    Display some current statistics on the repository
    """
    images = gcloud.list_images(gcr_repo)

    for image in images:
        tags = gcloud.list_tags(image)

        image_name = typer.style(
            image.removeprefix(gcr_repo + "/"), fg=typer.colors.BLUE
        )

        if not tags:
            ok = typer.style("EMPTY", fg=typer.colors.GREEN)
            typer.echo(f"{image_name} | {ok} | No tags found, nothing to do.")
            continue

        oldest_ts = None
        oldest_tag = None
        latest_ts = None
        latest_tag = None

        for tag in tags:
            if oldest_ts is None or tag.timestamp < oldest_ts:
                oldest_ts = tag.timestamp
                oldest_tag = tag.name
            if latest_ts is None or tag.timestamp > latest_ts:
                latest_ts = tag.timestamp
                latest_tag = tag.name

        keep, clean = _pick_cleanup_tags(tags)

        count = typer.style(f"{len(clean)}/{len(tags)}", fg=typer.colors.CYAN)
        oldest = typer.style(
            f"{oldest_ts.strftime(DT_FORMAT)} ({oldest_tag})", fg=typer.colors.MAGENTA
        )
        latest = typer.style(
            f"{latest_ts.strftime(DT_FORMAT)} ({latest_tag})", fg=typer.colors.MAGENTA
        )
        ok = (
            typer.style("NEEDS CLEANUP", fg=typer.colors.BRIGHT_RED)
            if clean
            else typer.style("OK", fg=typer.colors.GREEN)
        )

        typer.echo(
            f"{image_name} | {ok} | {count} images to clean | {oldest} - {latest}"
        )


@app.command()
def cleanup(gcr_repo: str):
    """
    Clean up some of the unneeded images
    """
    images = gcloud.list_images(gcr_repo)
    for image in images:
        typer.echo(typer.style(image, fg=typer.colors.BRIGHT_CYAN, bold=True))
        typer.echo(
            typer.style("-" * len(image), fg=typer.colors.BRIGHT_CYAN, bold=True)
        )
        typer.echo("")

        tags = gcloud.list_tags(image)
        keep, clean = _pick_cleanup_tags(tags)

        for tag in clean:
            typer.echo(
                typer.style(
                    f"Removing {', '.join(tag.tags)} from {tag.timestamp.strftime(DT_FORMAT)}",
                    fg=typer.colors.BRIGHT_YELLOW,
                )
            )
            gcloud.delete_tag(image, tag)

        if clean:
            typer.echo("")

        for tag in keep:
            typer.echo(
                typer.style(
                    f"Keeping {tag.name} from {tag.timestamp.strftime(DT_FORMAT)}",
                    fg=typer.colors.BRIGHT_GREEN,
                )
            )

        typer.echo("")



import json
from collections import namedtuple
from datetime import datetime

import pydantic
import requests
from requests.auth import HTTPBasicAuth

from gcrc.utils import run

_AUTH_TOKEN: str = ""
ImageRow = namedtuple("ImageRow", ["name"])


class ImageTag(pydantic.BaseModel):
    digest: str
    name: str
    tags: list[str]
    timestamp: datetime

    @classmethod
    def pre_process(cls, data):
        try:
            name = data["tags"][0]
        except IndexError:
            name = data["digest"]

        return {
            "digest": data["digest"],
            "name": name,
            "tags": data["tags"],
            "timestamp": data["timestamp"]["datetime"],
        }


def get_auth_token() -> str:
    global _AUTH_TOKEN
    if not _AUTH_TOKEN:
        _AUTH_TOKEN = run(["gcloud", "auth" "print-access-token"]).stdout.decode(
            "utf-8"
        )

    return _AUTH_TOKEN


def list_images(repository: str) -> list[str]:
    cmd = [
        "gcloud",
        "container",
        "images",
        "list",
        "--format",
        "json",
        "--repository",
        repository,
    ]
    result = run(cmd)
    return [ImageRow(**row).name for row in json.loads(result.stdout)]


def list_tags(image: str) -> list[ImageTag]:
    cmd = ["gcloud", "container", "images", "list-tags", "--format", "json", image]
    result = run(cmd)
    return [ImageTag(**ImageTag.pre_process(row)) for row in json.loads(result.stdout)]


def delete_tag(image: str, tag: ImageTag):
    name = f"{image}@{tag.digest}"
    cmd = [
        "gcloud",
        "container",
        "images",
        "delete",
        "--quiet",
        "--force-delete-tags",
        name,
    ]
    run(cmd)


def delete_manifest(repository: str, image: str, tag: ImageTag):
    parts = repository.split("/", maxsplit=2)
    registry = parts[0]
    auth = HTTPBasicAuth("_token", get_auth_token())
    url = f"https://{registry}/v2/{image}/manifests/{tag.digest}"
    requests.delete(url, auth=auth)



from pydantic import BaseSettings


class Settings(BaseSettings):
    # Keep at minimum this many tags
    KEEP_TAGS_MIN: int = 10

    # Only clean up things older than this many days
    KEEP_TAGS_DAYS: int = 14

    # Keep more `KEEP_TAGS_MIN` instances of any tags matching these regexes
    KEEP_EXTRA: list[str] = ["^latest$", "^(master|main)-"]

    class Config:
        env_file = ".env"
        env_file_encoding = "utf-8"


conf = Settings()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gcrc']

package_data = \
{'': ['*']}

install_requires = \
['colorama>=0.4.4,<0.5.0',
 'pydantic[dotenv]>=1.8.2,<2.0.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'requests>=2.26.0,<3.0.0',
 'typer[colorama]>=0.3.2,<0.4.0']

entry_points = \
{'console_scripts': ['gcrc = gcrc.main:app']}

setup_kwargs = {
    'name': 'gcrc',
    'version': '1.0.6',
    'description': 'Google Container Registry Cleanup utility',
    'long_description': '[![GitHub Workflow Status](https://img.shields.io/github/workflow/status/lietu/gcrc/Build%20and%20upload%20to%20PyPI)](https://github.com/lietu/gcrc/actions/workflows/build-and-upload.yaml)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![PyPI](https://img.shields.io/pypi/v/gcrc)](https://pypi.org/project/gcrc/)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/gcrc)](https://pypi.org/project/gcrc/)\n[![License: BSD 3-Clause](https://img.shields.io/badge/License-BSD%203--Clause-blue.svg)](https://opensource.org/licenses/BSD-3-Clause)\n\n# Google Container Registry Cleanup utility\n\nIf you use a CI system to push images to Google Container Registry you will eventually run into a situation where you\'re paying more than you would like for the storage there.\n\nThis tool helps you keep those costs reasonable, while allowing you to benefit from caches, keeping backups for rollbacks, etc.\n\n## Prerequisites and setup\n\n- [Python 3.9+](https://www.python.org/downloads/)\n- [Gcloud SDK](https://cloud.google.com/sdk/docs/install) with an authentication token that has admin access to the registry properly configured\n\n```bash\npip install -U gcrc\n```\n\n## Usage\n\nReplace `gcr.io/project-name` with the specific GCR address to your repository, e.g. if your Google Cloud project is called `foo-bar` and it\'s in the EU zone this would likely be `eu.gcr.io/foo-bar`.\n\n### List all images\n\nFigures out the different images we can access in the Container Registry.\n\n```bash\ngcrc list-images gcr.io/project-name\n```\n\n### Show information on images\n\nAnalyzes current images and any need for cleanup based on current configuration without actually deleting anything. Use to check that configuration seems correct.\n\n```bash\ngcrc image-info gcr.io/project-name\n```\n\n### Clean up images\n\nLook up images with unneeded tags and delete them from Google Container Registry\n\n```bash\ngcrc cleanup gcr.io/project-name\n```\n\n## Configuration\n\nThe following environment variables can be used to adjust the configuration:\n\n```\nKEEP_TAGS_MIN=10\n```\n\nKeeps at least this many tags for every image.\n\n```\nKEEP_TAGS_DAYS=14\n```\n\nKeep everything from within this many days.\n\n```\nKEEP_EXTRA=\'["^latest$", "^(master|main)-"]\'\n```\n\nList of regex matches for important images that we want to keep an extra `KEEP_TAGS_MIN` items of. Formatted as a JSON list of strings.\n\nThe default is to keep `latest` and any tags starting with `master-` or `main-` for people who tag their images like `<branch>-<timestamp or commit hash or uuid>`.\n\nYou can also put these in a `.env` -file, but environment variables take priority over `.env`.\n\n```\n# .env\nKEEP_TAGS_MIN=10\nKEEP_TAGS_DAYS=14\nKEEP_EXTRA=\'["^important-", "^latest$"]\'\n```\n\n# Financial support\n\nThis project has been made possible thanks to [Cocreators](https://cocreators.ee) and [Lietu](https://lietu.net). You can help us continue our open source work by supporting us on [Buy me a coffee](https://www.buymeacoffee.com/cocreators).\n\n[!["Buy Me A Coffee"](https://www.buymeacoffee.com/assets/img/custom_images/orange_img.png)](https://www.buymeacoffee.com/cocreators)\n',
    'author': 'Janne Enberg',
    'author_email': 'janne.enberg@lietu.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/lietu/gcrc',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
